use std::collections::{HashMap, HashSet};

use crate::expr::{Expr, FlattenedExpr};

pub struct Flattener {
    funcs: HashMap<String, Expr>,
}

impl Flattener {
    pub fn new(funcs: HashMap<String, Expr>) -> Self {
        Self { funcs }
    }

    pub fn flatten(&mut self, main: &Expr) -> Option<FlattenedExpr> {
        self.flatten_expr(main, HashSet::new())
    }

    fn flatten_expr<'a>(
        &self,
        expr: &'a Expr,
        mut names: HashSet<&'a str>,
    ) -> Option<FlattenedExpr> {
        let ret = match expr {
            Expr::Abs(expr) => FlattenedExpr::Abs(Box::new(self.flatten_expr(expr, names)?)),
            Expr::App(left, right) => FlattenedExpr::App(
                Box::new(self.flatten_expr(left, names.clone())?),
                Box::new(self.flatten_expr(right, names)?),
            ),
            Expr::Var(x) => FlattenedExpr::Var(*x),
            Expr::DefinitionName(name) => {
                if names.contains(name.as_str()) {
                    self.print_error(&format!("Recursive definition for ${}", name));
                    return None;
                }

                let expr = match self.funcs.get(name) {
                    Some(x) => x,
                    None => {
                        self.print_error(&format!("Unknown definition ${}", name));

                        return None;
                    }
                };

                names.insert(name);

                self.flatten_expr(expr, names)?
            }
        };

        Some(ret)
    }

    fn print_error(&self, err: &str) {
        eprintln!("Error: {err}");
    }
}
